#include "sip_util.h"
#include "sip_ctrl.h"
#include "sip_message.h"
#include "sdp.h"
#include "stack_if.h"

#ifdef _WINDOWS
#include<time.h>
#endif

extern SAVE_MSG   g_SipMsg;
extern STACK_INFO g_StackInfo;

/* ¸Ƥåؤ */
void save_message(char* msg, SAVE_MSG* s_msg){
    strcpy(s_msg->save_beforMsg, s_msg->save_Msg);
    strcpy(s_msg->save_Msg,msg);
}

/* ϿѤREGISTER */
int SendRegister(char* msg, char* username, char* serv_addr){
    int ret;
    char hostname[MAX_NAME];
    char* addrbuff;
    struct hostent* phost;
    struct hostent* psvrhost;
    struct in_addr iaddr;

    if(username == NULL){
        debug("Error!!SendRegister failed.\n");
        return -1;
    }
    g_StackInfo.state = eSRegister;
    g_StackInfo.seq = 1+RANDOM(10000);
    gethostname(hostname,MAX_NAME);
    phost = gethostbyname(hostname);

    /* üΥ桼̾IPɥ쥹FQDN */
    strcpy(g_StackInfo.fqdn, phost->h_name);
    addrbuff = inet_ntoa(*(struct in_addr*)phost->h_addr_list[0]);
    strcpy(g_StackInfo.addr, addrbuff);

    /* ФIPɥ쥹FQDN */
    inet_aton(serv_addr, &iaddr);
    psvrhost = gethostbyaddr((char*)&iaddr, 4, AF_INET);
    if(psvrhost == NULL){
        debug("SIP server name is not resolved by addr.\n");
        return -1;
    }
    strcpy(g_StackInfo.serv_fqdn, psvrhost->h_name);

    /* ȥ饤 */
    ret = set_Startline(Method_Register, msg, g_StackInfo.serv_fqdn);
    if(ret<0){
        debug("Error!!SendRegister set_Startline failed.\n");
        return -1;
    }

    /* Via */
    {
        char field[MAX_FIELD];
        Make_Via(field, g_StackInfo.addr, g_StackInfo.seq);
        ret = set_Via(msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_Via failed.\n");
            return -1;
        }
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendRegister set_Maxforwards failed.\n");
        return -1;
    }

    /* FromTo */
    {
        char field[MAX_FIELD];
        char fieldto[MAX_FIELD];
        sprintf(field, "%s <sip:%s@%s>;tag=s%d" ,
         g_StackInfo.username, g_StackInfo.username, 
         g_StackInfo.fqdn, RANDOM(100000));
        ret = set_From( msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_From failed.\n");
            return -1;
        }
        sprintf(fieldto, "%s <sip:%s@%s>" , g_StackInfo.username, 
         g_StackInfo.username, g_StackInfo.fqdn);
        ret = set_To( msg, fieldto);
        if(ret < 0) {
            debug("Error!!SendRegister set_To failed.\n");
            return -1;
        }
    }

    /* Call-ID */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d-%d-%d@%s",
                RANDOM(10000), RANDOM(10000), RANDOM(10000), g_StackInfo.fqdn);
        ret = set_CallID( msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_CallID failed.\n");
            return -1;
        }
    }

    /* CSeq */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d %s" , g_StackInfo.seq, METHOD_REGISTER);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_CallID failed.\n");
            return -1;
        }
    }

    /* Expires */
    ret = set_Expires(msg, MAX_EXPIRES);
    if(ret<0){
        debug("Error!!SendRegister set_Expires failed.\n");
        return -1;
    }

    /* Contact */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        sprintf(field, "%s@%s", g_StackInfo.username, g_StackInfo.addr);
        ret = set_Contact(msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_Contact failed.\n");
            return -1;
        }
    }

    /* Content-Length */
    {
        ret = set_Contentlength(msg, 0);
    }

    printf("\n<----- send REGISTER request\n");
    SendRegisterMsg(msg, g_StackInfo.serv_addr);

    /* ֹ1Ŀʤ */
    g_StackInfo.seq++;

    return 0;
}

/* ϿѤREGISTER */
int SendUnRegister(char* msg){
    int ret;

    /* ȥ饤 */
    ret = set_Startline(Method_Register, msg, g_StackInfo.serv_fqdn);
    if(ret<0){
        debug("Error!!SendUnRegister set_Startline failed.\n");
        return -1;
    }

    /* Via */
    {
        char field[MAX_FIELD];
        Make_Via(field, g_StackInfo.addr, g_StackInfo.seq);
        ret = set_Via(msg, field);
        if(ret<0){
            debug("Error!!SendUnRegister set_Via failed.\n");
            return -1;
        }
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendUnRegister set_Maxforwards failed.\n");
        return -1;
    }

    /* FromTo */
    {
        char field[MAX_FIELD];
        char fieldto[MAX_FIELD];
        sprintf(field, "%s <sip:%s@%s>;tag=s%d" ,
                g_StackInfo.username, g_StackInfo.username, 
                g_StackInfo.fqdn, RANDOM(100000));
        ret = set_From( msg, field);
        if(ret<0){
            debug("Error!!SendRegister set_From failed.\n");
            return -1;
        }
        sprintf(fieldto, "%s <sip:%s@%s>" , g_StackInfo.username, 
                g_StackInfo.username, g_StackInfo.fqdn);
        ret = set_To( msg, fieldto);
        if(ret<0){
            debug("Error!!SendUnRegister set_To failed.\n");
            return -1;
        }
    }

    /* Call-ID */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d-%d-%d@%s",
                RANDOM(10000), RANDOM(10000), RANDOM(10000), g_StackInfo.fqdn);
        ret = set_CallID( msg, field);
        if(ret<0){
            debug("Error!!SendUnRegister set_CallID failed.\n");
            return -1;
        }
    }

    /* CSeq */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d %s" , g_StackInfo.seq, METHOD_REGISTER);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendUnRegister set_CallID failed.\n");
            return -1;
        }
    }

    /* Expires */
    ret = set_Expires(msg, 0);
    if(ret<0){
        debug("Error!!SendUnRegister set_Expires failed.\n");
        return -1;
    }

    /* Contact */
    ret = set_ContactA(msg, g_StackInfo.addr, g_StackInfo.username);
    if(ret<0){
        debug("Error!!SendUnRegister set_Contact failed.\n");
        return -1;
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendUnRegister set_Contentlength failed.\n");
        return -1;
    }
    debug("\n<----- send REGISTER request(Expires = 0)\n");
    SendRegisterMsg(msg, g_StackInfo.serv_addr);
    return 0;
}

/* INVITE */
int SendInvite(char* msg, char* username, char* called, int rtpport){
    int ret;
    char hostname[MAX_NAME];
    struct hostent host;
    struct hostent* phost = &host;
    char*   addrbuff;

    /* SIPФλѡ̤ѤȽ */
    if(g_StackInfo.server_flag == eRNull){
        gethostname(hostname, MAX_NAME);
        phost = gethostbyname(hostname);

        /* ֹ¸ */
        g_StackInfo.seq = 1+RANDOM(10000);

        /* üΥ桼̾IPɥ쥹FQDN¸ */
        strcpy(g_StackInfo.fqdn, phost->h_name);
        addrbuff = inet_ntoa(*(struct in_addr*)phost->h_addr_list[0]);
        strcpy(g_StackInfo.addr, addrbuff);
    }
    if(g_StackInfo.username == NULL|| called == NULL){
        debug("Error!!SendInvite failed.\n");
        return -1;
    }

    /* ȥ饤 */
    ret = set_Startline(Method_Invite, msg, called);
    if(ret<0){
        debug("Error!!SendInvite set_Startline failed.\n");
        return -1;
    }

    /* Via */
    {
        char field[MAX_FIELD];
        Make_Via(field, g_StackInfo.addr, g_StackInfo.seq);
        ret = set_Via(msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_Via failed.\n");
            return -1;
        }
        /* Viaإåξ¸ */
        strcpy(g_SipMsg.save_Via, field);
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendInvite set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    {
        char field[MAX_FIELD];
        sprintf(field, "%s <sip:%s@%s>;tag=s%d" ,
                username, username, g_StackInfo.fqdn, RANDOM(100000));
        ret = set_From( msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_From failed.\n");
            return -1;
        }
        /* Fromإåξ¸ */
        strcpy(g_SipMsg.save_From, field);
    }

    /* To */
    {
        char* puser;
        char tmp[MAX_FIELD];
        char field[MAX_FIELD];
        strcpy(tmp, called);
        puser = strstr(tmp, "@");
        if(puser==0){
            debug("Error!!SendInvite To no @ failed.\n");
            return -1;
        }
        *puser = 0;                                     /* NULL */
        sprintf(field, "%s <sip:%s>" , tmp, called);
        ret = set_To( msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_To failed.\n");
            return -1;
        }
        /* Toإåξ¸ */
        strcpy(g_SipMsg.save_To, field);
    }

    /* Call-ID */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d-%d-%d@%s",
                RANDOM(10000), RANDOM(10000), RANDOM(10000), g_StackInfo.fqdn);
        ret = set_CallID( msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_CallID failed.\n");
            return -1;
        }
        /* Call-IDإåξ¸ */
        strcpy(g_SipMsg.save_CallID, field);
    }

    /* CSeq */
    {
        char field[MAX_FIELD];
        sprintf(field, "%d %s" , g_StackInfo.seq, METHOD_INVITE);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_CSeq failed.\n");
            return -1;
        }
        /* CSeqξ¸ */
        strcpy(g_SipMsg.save_Cseq, field);
    }

    /* Contact */
    ret = set_ContactA(msg, g_StackInfo.addr, username);
    if(ret<0){
        debug("Error!!SendInvite set_Contact failed.\n");
        return -1;
    }

    /* Content-Type */
    ret = set_Contenttype(msg, CONTENT_TYPE);
    if(ret<0){
        debug("Error!!SendInvite set_Contenttype failed.\n");
        return -1;
    }

    /* Content-LengthSDP */
    {
        int contentlen;
        char field[MAX_FIELD];
        make_SDP(field, g_StackInfo.username, g_StackInfo.addr,
                 g_StackInfo.fqdn, CODEC_AUDIO, RTPTYPE, AUDIO_RATE, g_StackInfo.rtpport, NULL);
        contentlen = strlen(field);
        ret = set_Contentlength(msg, contentlen);
        if(ret<0){
            debug("Error!!SendInvite set_Contentlength failed.\n");
            return -1;
        }
        set_SDP(msg, field);
    }

    /* ֹ1Ŀʤ */
    g_StackInfo.seq++;

    /* ȯԥե饰ΩƤ */
    g_StackInfo.session_flag = eCSender;

    /* å¸ */
    save_message(msg, &g_SipMsg);
    debug("\n<---- send INVITE request\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed in SendInvite.\n");
    }
    g_StackInfo.state = eSWaitBusy;
    return 0;
}

/* 100 Trying */
int SendTrying(char* msg){
    int ret;
    char field[MAX_FIELD];

    /* ȥ饤 */
    ret = set_Startline(Method_Trying, msg, NULL);
    if(ret<0){
        debug("Error!!SendTrying set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendTrying set_Via failed.\n");
        return -1;
    }
    if (g_StackInfo.server_flag == eRRegister){
        ret = set_Via(msg, g_SipMsg.save_Via2nd);
        if(ret<0){
            debug("Error!!SendTrying 2nd set_Via failed.\n");
            return -1;
        }
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendTrying set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendTrying set_From failed.\n");
        return -1;
    }

    /* To */
    /* Toʤϡɲä */
    strcpy(field, g_SipMsg.save_To);
    if(!strstr(g_SipMsg.save_To,"tag=")){
        char* mp = strchr(field, '\0');
        sprintf(mp, ";tag=r%d", RANDOM(100000));
    }
    strcpy(g_SipMsg.save_To, field);
    ret = set_To( msg, field);
    if(ret<0){
        debug("Error!!SendTrying set_To failed.\n");
        return -1;
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendTrying set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    ret = set_CSeq( msg, g_SipMsg.save_Cseq);
    if(ret<0){
        debug("Error!!SendTrying set_CSeq failed.\n");
        return -1;
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendTrying set_Contentlength failed.\n");
        return -1;
    }
    debug("\n<----- send Trying response\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* 180 Ringing */
int SendRinging(char* msg){
    int ret;

    /* ȥ饤 */
    ret = set_Startline(Method_Ringing, msg, NULL);
    if(ret<0){
        debug("Error!!SendRinging set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendRinging set_Via failed.\n");
        return -1;
    }
    if (g_StackInfo.server_flag == eRRegister){
        ret = set_Via(msg, g_SipMsg.save_Via2nd);
        if(ret<0){
            debug("Error!!SendTrying 2nd set_Via failed.\n");
            return -1;
        }
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendRinging set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendRinging set_From failed.\n");
        return -1;
    }

    /* To */
    ret = set_To( msg, g_SipMsg.save_To);
    if(ret<0){
        debug("Error!!SendRinging set_To failed.\n");
        return -1;
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendRinging set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    ret = set_CSeq( msg, g_SipMsg.save_Cseq);
    if(ret<0){
        debug("Error!!SendRinging set_CSeq failed.\n");
        return -1;
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendRinging set_Contentlength failed.\n");
        return -1;
    }

    ret = SendSipMessage(msg);
    debug("\n<----- send Ringing response\n");
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* 200 OK */
int SendOK(char* msg){
    int ret;
    enum method emethod;

    /* 쥹ݥ󥹤֤åμ̤Ƚꤹ */
    emethod = Search_Method(g_SipMsg.save_Msg);

    /* ȥ饤 */
    ret = set_Startline(Method_OK, msg, NULL);
    if(ret<0){
        debug("Error!!SendOK set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendOK set_Via failed.\n");
        return -1;
    }
    if(g_StackInfo.server_flag == eRRegister){
        ret = set_Via(msg, g_SipMsg.save_Via2nd);
    }else{
        *(g_SipMsg.save_Via2nd) = '\0'; 
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendInvite set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendInvite set_From failed.\n");
        return -1;
    }

    /* To */
    {
        char field[MAX_FIELD];
        char* mp;
        
        memset(field, 0, MAX_FIELD);
        strcpy(field, g_SipMsg.save_To);
        /* Toʤ硢ɲä */
        if(!strstr(g_SipMsg.save_To,"tag=")){
            mp = strchr(field, '\0');
            sprintf(mp, ";tag=r%d", RANDOM(100000));
        }
        ret = set_To( msg, field);
        if(ret<0){
            debug("Error!!SendInvite set_To failed.\n");
            return -1;
        }
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendInvite set_CallID failed.\n");
        return -1;
    }
    switch(g_StackInfo.state){
        case eSWaitBusy:{

            /* CSeq */
            ret = set_CSeq( msg, g_SipMsg.save_Cseq);
            if(ret<0){
                debug("Error!!SendInvite set_CSeq failed.\n");
                return -1;
            }

            /* Contact */
            ret = set_ContactA(msg, g_StackInfo.addr,
                g_StackInfo.username);
            if(ret<0){
                debug("Error!!SendInvite set_Contact failed.\n");
                return -1;
            }

            /* Content-Type */
            ret = set_Contenttype(msg, CONTENT_TYPE);
            if(ret<0){
                debug("Error!!SendInvite set_Contenttype failed.\n");
                return -1;
            }

            /* Content-Length, SDP */
            {
                int contentlen;
                char o[MAX_FIELD];
                char field[MAX_FIELD];
                memset(field, 0, MAX_FIELD);
                memset(o, 0, MAX_FIELD);
                /* INVITESDPoξ */
                get_o(g_SipMsg.save_Msg, o);
                make_SDP(field, g_StackInfo.username, g_StackInfo.addr,
                g_StackInfo.fqdn, CODEC_AUDIO, RTPTYPE, AUDIO_RATE, g_StackInfo.rtpport, o);
                contentlen = strlen(field);
                ret = set_Contentlength(msg, contentlen);
                if(ret<0){
                    debug("Error!!SendInvite set_Contentlength failed.\n");
                    return -1;
                }
                printf("ok\n");
                set_SDP(msg, field);
            }
        }break;
        case eSWaitDisc:{

            /* CSeq */
            ret = set_CSeq( msg, g_SipMsg.save_Cseq);
            if(ret<0){
                debug("Error!!SendInvite set_CSeq failed.\n");
                return -1;
            }

            /* Content-Length */
            ret = set_Contentlength(msg, 0);
            if(ret<0){
                debug("Error!!SendInvite set_Contentlength failed.\n");
                return -1;
            }
        }break;
        default:
            break;
    }
    debug("\n<----- send OK response\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* ACK */
int SendAck(char* msg){
    int ret;
    char* sp, *ep;
    char tmp[256], tmp2[256];

    /* ȥ饤 */
    strcpy(tmp, g_SipMsg.save_To);
    sp = strstr(tmp, "<sip:");
    ep = strstr(tmp, ">");
    strncpy(tmp2, sp+5, ep-(sp+5));
    tmp2[ep-(sp+5)] = '\0';
    ret = set_Startline(Method_Ack, msg, tmp2);
    if(ret<0){
        debug("Error!!SendAck set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendAck set_Via failed.\n");
        return -1;
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendAck set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendAck set_From failed.\n");
        return -1;
    }

    /* To */
    ret = set_To( msg, g_SipMsg.save_To);
    if(ret<0){
        debug("Error!!SendAck set_To failed.\n");
        return -1;
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendAck set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    {
        int seqnum;
        char field[MAX_FIELD];

        seqnum = get_CSeqnum(g_SipMsg.save_Cseq);
        memset(field, 0, MAX_FIELD);
        sprintf(field, "%d %s", seqnum, METHOD_ACK);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendAck set_CSeq failed.\n");
            return -1;
        }
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendAck set_Contentlength failed.\n");
        return -1;
    }
    debug("\n<----- send Ack request\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* BYE */
int SendBye(char* msg){
    int ret;

    /* ȥ饤 */
    {
        char buf[MAX_FIELD], buf2[MAX_FIELD], *ptr, *ptr2;
        if(g_StackInfo.session_flag == eCSender){
            strcpy(buf, g_SipMsg.save_To);
        }else{
            strcpy(buf, g_SipMsg.save_From);
        }
        ptr = strstr(buf, "sip:");
        ptr += 4;
        ptr2 = strstr(ptr, ">");
        *ptr2=0;
        strcpy(buf2, ptr);
        ret = set_Startline(Method_Bye, msg, buf2);
        if(ret<0){
            debug("Error!!SendBye set_Startline failed.\n");
            return -1;
        }
    }

    /* Via */
    {
        char field[MAX_FIELD];

        memset(field, 0, MAX_FIELD);
        Make_Via(field, g_StackInfo.addr, g_StackInfo.seq);
        ret = set_Via(msg, field);
        if(ret<0){
            debug("Error!!SendBye set_Via failed.\n");
            return -1;
        }

        /* Viaإåξ¸ */
        strcpy(g_SipMsg.save_Via, field);
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendBye set_Maxforwards failed.\n");
        return -1;
    }

    if(g_StackInfo.session_flag == eCSender){

        /* From */
        ret = set_From( msg, g_SipMsg.save_From);
        if(ret<0){
            debug("Error!!SendBye set_From failed.\n");
            return -1;
        }
        
        /* To */
        ret = set_To( msg, g_SipMsg.save_To);
        if(ret<0){
            debug("Error!!SendBye set_To failed.\n");
            return -1;
        }
    }else {

        /* From */
        ret = set_From( msg, g_SipMsg.save_To);
        if(ret<0){
            debug("Error!!SendBye set_From failed.\n");
            return -1;
        }
        
        /* To */
        ret = set_To( msg, g_SipMsg.save_From);
        if(ret<0){
            debug("Error!!SendBye set_To failed.\n");
            return -1;
        }
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendBye set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        sprintf(field, "%d %s", g_StackInfo.seq, METHOD_BYE);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendBye set_CSeq failed.\n");
            return -1;
        }
    }

    /* Contact */
    ret = set_ContactA(msg, g_StackInfo.fqdn, g_StackInfo.username);
    if(ret<0){
        debug("Error!!SendBye set_Contact failed.\n");
        return -1;
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendBye set_Contentlength failed.\n");
        return -1;
    }
    save_message(msg, &g_SipMsg);
    debug("\n<----- send BYE request\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
        return -1;
    }
    g_StackInfo.state = eSWaitDisc;
    return 0;
}

/* CANCEL */
int SendCancel(char* msg){
    int ret;
    char* sp, *ep;
    char tmp[256], tmp2[256];

    /* ȥ饤 */
    strcpy(tmp, g_SipMsg.save_To);
    sp = strstr(tmp, "<sip:");
    ep = strstr(tmp, ">");
    strncpy(tmp2, sp+5, ep-(sp+5));
    tmp2[ep-(sp+5)] = '\0';
    ret = set_Startline(Method_Cancel, msg, tmp2);
    if(ret<0){
        debug("Error!!SendCancel set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendCancel set_Via failed.\n");
        return -1;
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendCancel set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendCancel set_From failed.\n");
        return -1;
    }

    /* To */
    ret = set_To( msg, g_SipMsg.save_To);
    if(ret<0){
        debug("Error!!SendCancel set_To failed.\n");
        return -1;
    }
    
    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendCancel set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    {
        int seqnum;
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        seqnum = get_CSeqnum(g_SipMsg.save_Cseq);
        sprintf(field, "%d %s", seqnum, METHOD_CANCEL);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendCancel set_CSeq failed.\n");
            return -1;
        }
    }

    /* Contact */
    ret = set_ContactA(msg, g_StackInfo.addr, g_StackInfo.username);
    if(ret<0){
        debug("Error!!SendCancel set_Contact failed.\n");
        return -1;
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendCancel set_Contentlength failed.\n");
        return -1;
    }
    save_message(msg, &g_SipMsg);
    printf("\n<----- send CANCELL request\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
        return -1;
    }
    g_StackInfo.state = eSWaitDisc;
    return 0;
}

/* 486 Busy Here */
int SendBusy(char* msg, char* r_msg){
    int ret;

    /* ȥ饤 */
    ret = set_Startline(Method_Busy, msg, NULL);
    if(ret<0){
        debug("Error!!SendBusy set_Startline failed.\n");
        return -1;
    }

    /* Via */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        get_Via(r_msg, field);
        ret = set_Via(msg, field);
        if(g_StackInfo.server_flag == eRRegister){
            get_Via2nd(r_msg, field);
            ret = set_Via(msg, field);
        }
        if(ret<0){
            debug("Error!!SendBusy set_Via failed.\n");
            return -1;
        }
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendBusy set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        get_From(r_msg, field);
        ret = set_From( msg, field);
        if(ret<0){
            debug("Error!!SendBusy set_From failed.\n");
            return -1;
        }
    }

    /* To */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        get_To(r_msg, field);
        ret = set_To( msg, field);
        if(ret<0){
            debug("Error!!SendBusy set_To failed.\n");
            return -1;
        }
    }

    /* Call-ID */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        get_CallID(r_msg, field);
        ret = set_CallID( msg, field);
        if(ret<0){
            debug("Error!!SendBusy set_CallID failed.\n");
            return -1;
        }
    }

    /* CSeq */
    {
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        get_CSeq( r_msg, field);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendAck set_CSeq failed.\n");
            return -1;
        }
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendBusy set_Contentlength failed.\n");
        return -1;
    }

    debug("<----- send Busy Here response\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* 487 Request Terminate */
int SendRequestTerminate(char* msg){
    int ret;

    /* ȥ饤 */
    ret = set_Startline(Method_RequestTerminate, msg, NULL);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_Startline failed.\n");
        return -1;
    }

    /* Via */
    ret = set_Via(msg, g_SipMsg.save_Via);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_Via failed.\n");
        return -1;
    }

    /* Max-Forwards */
    ret = set_Maxforwards(msg, MAX_COUNT);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_Maxforwards failed.\n");
        return -1;
    }

    /* From */
    ret = set_From( msg, g_SipMsg.save_From);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_From failed.\n");
        return -1;
    }

    /* To */
    ret = set_To( msg, g_SipMsg.save_To);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_To failed.\n");
        return -1;
    }

    /* Call-ID */
    ret = set_CallID( msg, g_SipMsg.save_CallID);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_CallID failed.\n");
        return -1;
    }

    /* CSeq */
    {
        int seqnum;
        char field[MAX_FIELD];
        memset(field, 0, MAX_FIELD);
        seqnum = get_CSeqnum(g_SipMsg.save_Cseq);
        sprintf(field, "%d %s", seqnum, METHOD_INVITE);
        ret = set_CSeq( msg, field);
        if(ret<0){
            debug("Error!!SendRequestTerminate set_CSeq failed.\n");
            return -1;
        }
    }

    /* Content-Length */
    ret = set_Contentlength(msg, 0);
    if(ret<0){
        debug("Error!!SendRequestTerminate set_Contentlength failed.\n");
        return -1;
    }

    debug("<----- send Request Terminate response\n");
    ret = SendSipMessage(msg);
    if(ret < 0){
        debug("SendSipMessage failed.\n");
    }
    return 0;
}

/* ꥯȤμ̤Ϥ */
int RecvRequest(enum method request, char* msg, RECV_INFO* info){
    int ret;

    switch(request){
        /* INVITE */
        case Method_Invite:{
            debug("\n-----> received INVITE request\n");
            ret = RecvInvite(msg, info);
            if(ret<0){
                debug("RecvInvite failed\n");
                return -1;
            }
        }break;

        /* ACK */
        case Method_Ack:{
            debug("\n-----> received ACK request \n");
            ret = RecvAck(msg, info);
            if(ret < 0){
                debug("RecvAck failed\n");
                return -1;
            }
        }break;

        /* BYE */
        case Method_Bye:{
            debug("\n-----> received BYE request \n");
            ret = RecvBye(msg, info);
            if(ret<0){
                debug("RecvBye failed\n");
                return -1;
            }
        }break;

        /* CANCEL */
        case Method_Cancel:{
            debug("\n-----> received CANCEL request \n");
            ret = RecvCancel(msg, info);
            if(ret<0){
                debug("RecvCancel failed\n");
                return -1;
            }
        }break;

        /* ֹʤ */
        default:{
            debug("Error!!Receive request undefined.\n");
            return -1;
        }
    }
    return 0;
}

/* 쥹ݥ󥹤μ̤Ϥ */
int RecvResponse(enum method response, char* msg, RECV_INFO* info){
    int ret;

    switch(response){

        /* 100 Trying */
        case Method_Trying:{
            debug("\n-----> received Trying response\n");
        }break;

        /* 180 Ringing */
        case Method_Ringing:{
            debug("\n-----> received Ringing response\n");
        }break;

        /* 200 OK */
        case Method_OK:{
            debug("\n-----> received OK response\n");
            ret = RecvOK(msg, info);
            if(ret<0){
                debug("RecvOK failed\n");
                return -1;
            }
        }break;

        /* 486 Busy Here */
        case Method_Busy:{
            char ack[MAX_MESSAGE];
            debug("\n-----> received BusyHere response\n");
            ret = SendAck(ack);
            if(ret<0){
                debug("SendAck failed\n");
                return -1;
            }
            Finish(info->pmsg, info->pinfo);
        }break;

        /* 487 Request Terminate */
        case Method_RequestTerminate:{
            char ack[MAX_MESSAGE];
            debug("\n-----> received RequestTerminate response\n");
            ret = SendAck(ack);
            if(ret<0){
                debug("SendAck failed\n");
                return -1;
            }
            Finish(info->pmsg, info->pinfo);
        }break;

        /* ֹʤ */
        default:{
            debug("Error!!Receive responce undefined.\n");
            return -1;
        }
    }
    return 0;
}

/* INVITEνԤ */
int RecvInvite(char* msg, RECV_INFO* info){
    int ret;
    char line[MAX_FIELD];
    char trying[MAX_MESSAGE];
    char ringing[MAX_MESSAGE];
    char hostname[MAX_NAME];
    struct hostent host;
    struct hostent* phost = &host;
    char* addrbuff;

    /* ɥ֤Ǥʤ */
    if(info->pinfo->state != eSIdle){
        debug("Error!!Invalid state.\n");
        if(info->pinfo->state == eSBusy|| info->pinfo->state == eSWaitBusy){
            char busy[MAX_MESSAGE];
            ret = SendBusy(busy, msg);
            if(ret<0){
                debug("SendBusy failed.\n");
                return -1;
            }
            return 0;
        }
        return -1;
    }

    if(g_StackInfo.server_flag == eRNull){
        gethostname(hostname, MAX_NAME);
        phost = gethostbyname(hostname);
        
        strcpy(info->pinfo->fqdn, phost->h_name);
        addrbuff = inet_ntoa(*(struct in_addr*)phost->h_addr_list[0]);
        strcpy(info->pinfo->addr, addrbuff);
    }

    /* ѹ */
    info->pinfo->session_flag = eCReceiver;
    info->pinfo->state = eSWaitBusy;    

    /* åǤȤ¸ */
    /* INVITEå¸ */
    save_message(msg, info->pmsg);
    memset(line, 0, MAX_FIELD);

    /* Viaإåξ */
    get_Via(msg, line);
    strcpy(info->pmsg->save_Via, line);
    if(g_StackInfo.server_flag == eRRegister){
        get_Via2nd(msg, line);
        strcpy(info->pmsg->save_Via2nd, line);
    }else{
        *(info->pmsg->save_Via2nd) = '\0';  
    }
    memset(line, 0, MAX_FIELD);
    /* Fromإåξ */
    get_From(msg, line);
    strcpy(info->pmsg->save_From, line);
    memset(line, 0, MAX_FIELD);
    /* Toإåξ */
    get_To(msg, line);
    strcpy(info->pmsg->save_To, line);
    memset(line, 0, MAX_FIELD);
    /* Call-IDξ */
    get_CallID(msg, line);
    strcpy(info->pmsg->save_CallID, line);
    memset(line, 0, MAX_FIELD);
    /* CSeqξ */
    get_CSeq(msg, line);
    strcpy(info->pmsg->save_Cseq, line);

    get_rtpaddr(msg, info->pinfo->rtp_addr);
    get_rtpport(msg, info->pinfo->y_rtpport);{
        char buf[MAX_FIELD];
        sprintf(buf, "RTP destination(calling) IP addr is %s\n", info->pinfo->rtp_addr);
        debug(buf);
        sprintf(buf, "RTP destination(calling) port is %s\n", 
            info->pinfo->y_rtpport);
        debug(buf);
    }
    ret = SendTrying(trying);   /* 100 Trying*/
    if(ret<0){
        debug("SendTrying failed\n");
        return -1;
    }

    ret = SendRinging(ringing); /* 180 Ringing */
    if(ret<0){
        debug("SendRinging failed.\n");
        return -1;
    }
    incomingcall();
    return 0;
}

/* 200 OKνԤ */
int RecvOK(char* msg, RECV_INFO* info){
    int ret;
    
    switch(info->pinfo->state){
        case eSRegister:{
            info->pinfo->state = eSIdle;
        }break;
        case eSWaitBusy:{
            char callid[MAX_FIELD];
            char line[MAX_FIELD];
            char ack[MAX_MESSAGE];

            /* üȯԤǤ뤳Ȥǧ */
            if(info->pinfo->session_flag != eCSender){
                debug("Error!!Invalid state.\n");
                return -1;
            }
            get_CSeq(msg, callid);

            /* CSeqɤӤ */
            if(0 != strcmp(info->pmsg->save_Cseq, callid)){
                debug("Receive message's CSeq is different saved one.\n");
                return -1;
            }

            /* SIPåξǤ¸ */
            get_To(msg, line);

            /* ¸ƤToξ */
            memset(info->pmsg->save_To, 0, MAX_FIELD);

            /* OKToξ¸ */
            strcpy(info->pmsg->save_To, line);

            get_rtpaddr(msg, info->pinfo->rtp_addr);
            get_rtpport(msg, info->pinfo->y_rtpport);

            ret = SendAck(ack);
            if(ret<0){
                debug("SendAck failed\n");
                return -1;
            }
            info->pinfo->state = eSBusy;
            connection();
        }break;
        case eSWaitDisc:{
            enum method emethod;
            emethod = Search_Method(info->pmsg->save_Msg);
            if(emethod == Method_Bye){

                /* Ʋ */
                disconnect();
                Finish(info->pmsg, info->pinfo);
            }
        }break;
        case eSIdle:{
            printf("Unregister done.\n");
        }break;
        default:{
            printf("eSWaitDisc %d: state %d\n", eSWaitDisc, info->pinfo->state);
            debug("Error!!Invalid state.\n");
            return -1;
        }
    }
    return 0;
}

/* ACKνԤ */
int RecvAck(char* msg, RECV_INFO* info){
    char cid[MAX_FIELD];

    /* Call-IDå */
    get_CallID( msg, cid);

    /* Call-ID㤦 */
    if(0 != strcmp(cid, info->pmsg->save_CallID)){
        /* 486ФACK⤷ʤΤǲ⤷ʤ */
        return 0;
    }

    /* ü忮ԤǤ뤳Ȥǧ */
    if(info->pinfo->session_flag != eCReceiver){
        debug("Error!!Invalid state.\n");
        return -1;
    }
    switch(info->pinfo->state){
        /* ³ξ硢ä򳫻Ϥ */
        case eSWaitBusy: info->pinfo->state = eSBusy;
            printf("if you send audio rtp, please input 's'.\n");
            printf("if you want to disconnect this call, please input 'b'.\n");
            break;
        /* ǽξ硢ƲԤ */
        case eSWaitDisc: disconnect();
            Finish(info->pmsg, info->pinfo);
            break;
        default: debug("Error!!Invalid state.\n");
            return -1;
            break;
    }
    return 0;
}

/* BYEνԤ */
int RecvBye(char* msg, RECV_INFO* info){
    int ret;
    char line[MAX_FIELD];
    char ok[MAX_MESSAGE];

    memset(ok, 0, MAX_MESSAGE);
    if(info->pinfo->state != eSBusy && info->pinfo->state != eSWaitBusy){
        debug("Error!!Invalid state.\n");
        return -1;
    }
    info->pinfo->state = eSWaitDisc;
    save_message(msg, info->pmsg);

    /* Viaإåξ */
    memset(line, 0, MAX_FIELD);
    memset(info->pmsg->save_Via, 0, MAX_FIELD);
    get_Via(msg, line);
    strcpy(info->pmsg->save_Via, line);
    if(g_StackInfo.server_flag == eRRegister){
        get_Via2nd(msg, line);
        strcpy(info->pmsg->save_Via2nd, line);
    }else{
        *(info->pmsg->save_Via2nd) = '\0';
    }

    /* CSeqξ */
    memset(line, 0, MAX_FIELD);
    memset(info->pmsg->save_Cseq, 0, MAX_FIELD);
    get_CSeq(msg, line);
    strcpy(info->pmsg->save_Cseq, line);

    *(info->psend_switch) = ng;

    ret = SendOK(ok);
    if(ret < 0){
        debug("SendOK failed\n");
        return -1;
    }
    disconnect();
    Finish(info->pmsg, info->pinfo);
    return 0;
}

/* CANCELνԤ */
int RecvCancel(char* msg, RECV_INFO* info){
    int ret;
    char line[MAX_FIELD];
    char ok[MAX_MESSAGE];
    char terminate[MAX_MESSAGE];

    if(info->pinfo->state != eSWaitBusy){
        debug("Error!!Invalid state.\n");
        return -1;
    }

    /* ü忮ԤǤ뤳Ȥǧ */
    if(info->pinfo->session_flag != eCReceiver){
        debug("Error!!Invalid state.\n");
        return -1;
    }
    info->pinfo->state = eSWaitDisc;
    memset(line, 0, MAX_FIELD);
    get_CSeq(msg, line);

    /* CSeqɤӤ */
    if(get_CSeqnum(line) != get_CSeqnum(info->pmsg->save_Cseq)){
        debug("Receive message's CSeq was invalid.\n");
        return -1;
    }
    
    info->pinfo->state = eSWaitDisc;

    /* å¸ */
    save_message(msg, info->pmsg);

    /* CSeqξ¸ */
    memset(info->pmsg->save_Cseq, 0, MAX_FIELD);
    strcpy(info->pmsg->save_Cseq, line);

    ret = SendOK(ok);
    if(ret < 0){
        debug("SendOK failed\n");
        return -1;
    }

    ret = SendRequestTerminate(terminate);
    if(ret < 0){
        debug("SendRequestTerminate failed\n");
        return -1;
    }
    return 0;
}

/* Viaإå */
void Make_Via(char* via, char* addr, int seq){
    int time;
    long timestamp;
    struct timeval tv;
    struct timezone tz;
    if (-1 == gettimeofday(&tv, &tz)){
        return;
    }else{
        time = tv.tv_sec*1000 + tv.tv_usec/1000;
    }
    timestamp = (long) (time & 0x7FFFFFFF);

    sprintf(via, "%s/%s %s:%d;branch=z9hG4bK-%d-%d", 
            SIP_VERSION, PROTOCOL_UDP, addr, SIP_PORT, timestamp, seq);
}

/* ƽλ¸Ƥ */
void Finish(SAVE_MSG* msg, STACK_INFO* s_info){
    /* ¸ν */
    memset(msg->save_Msg, 0, MAX_MESSAGE);
    memset(msg->save_beforMsg, 0, MAX_MESSAGE);
    memset(msg->save_Via, 0, MAX_FIELD);
    memset(msg->save_From, 0, MAX_FIELD);
    memset(msg->save_To, 0, MAX_FIELD);
    memset(msg->save_CallID, 0, MAX_FIELD);
    memset(msg->save_Cseq, 0, MAX_FIELD);

    s_info->session_flag = eCNull;
    s_info->state = eSIdle;
}
